import requests

def get_science_direct_pdf(doi, api_key):
    base_url = "https://api.elsevier.com/content/article/doi/"
    headers = {
        "X-ELS-APIKey": api_key,
        "Accept": "application/pdf"
    }

    response = requests.get(f"{base_url}{doi}", headers=headers)

    if response.status_code == 200:
        # Assuming the response is a PDF file
        with open(f"{doi.replace('/', '_')}.pdf", "wb") as f:
            f.write(response.content)
        return f"PDF saved as {doi.replace('/', '_')}.pdf"
    else:
        return "Failed to retrieve PDF: " + response.text

# Usage
your_doi = "10.1519/JSC.0b013e318225bbae"
your_api_key = "e009797d502ddbdeedfb744abd63f5d3"  # Replace with your Elsevier API key
print(get_science_direct_pdf(your_doi, your_api_key))


#!/usr/bin/env python3
from __future__ import annotations
import sys
from enum import Enum
from functools import partial
from multiprocessing import Pool, cpu_count
from pathlib import Path
from typing import Optional, List
import click
import loguru
import pynction.monads.try_monad
import tiktoken
from click import Context
from functional import seq
from langchain.document_loaders import UnstructuredPDFLoader, PDFMinerLoader, PyPDFLoader, PyMuPDFLoader, \
    PDFPlumberLoader
from langchain.schema import Document
from pycomfort.files import traverse, files
from pynction import Try
import loguru
from getpaper.download import try_download, PaperDownload

from pycomfort.config import configure_logger, LOG_LEVELS, LogLevel


class PDFParser(Enum):
    unstructured = "unstructured"
    pdf_miner = "pdf_miner"
    py_pdf = "py_pdf"
    py_mu_pdf = "py_mu_pdf"
    pdfplumber = "pdfplumber"


def num_tokens_openai(string: str, model: str, price_per_1k: float = 0.0001) -> (int, float):
    """Returns the number of tokens for a model"""
    encoding = tiktoken.encoding_for_model(model)
    n_tokens = len(encoding.encode(string))
    return n_tokens, n_tokens / 1000.0 * price_per_1k


"""
    Model	Input	Output
8K context	$0.03 / 1K tokens	$0.06 / 1K tokens
32K context	$0.06 / 1K tokens	$0.12 / 1K tokens
"""

openai_prices_per_thousand = {
    "gpt-4-32k": None,  # 32,768 tokens
    "gpt-4": 0.03,  # 8,192 tokens
    "gpt-3.5-turbo-16k": None,  # 16,384 tokens,
    "gpt-3.5-turbo": None  # 4,096 tokens
}


def clean_text(text: str) -> str:
    from unstructured.cleaners.core import clean, group_broken_paragraphs, replace_unicode_quotes
    return clean(group_broken_paragraphs(replace_unicode_quotes(text)))


def parse_paper(paper: Path, folder: Optional[Path] = None,
                parser: PDFParser = PDFParser.py_mu_pdf,
                recreate_parent: bool = False,
                subfolder: bool = True,
                do_not_reparse: bool = True,
                cleaning: bool = False,
                mode: str = "single", strategy: str = "auto",
                pdf_infer_table_structure: bool = True,
                include_page_breaks: bool = False,
                logger: Optional["loguru.Logger"] = None
                ) -> List[Path]:
    """
    :param paper:
    :param folder:
    :param parser:
    :param recreate_parent:
    :param cleaning:
    :param mode: unstructured specific
    :param strategy:  unstructured specific
    :param pdf_infer_table_structure:  unstructured specific
    :param include_page_breaks:  unstructured specific
    :return:
    """
    if logger is None:
        logger = loguru.logger
    if parser is None or parser == PDFParser.unstructured:
        loader = init_unstructured_loader(open(str(paper), "rb"), include_page_breaks, mode, pdf_infer_table_structure,
                                          strategy)
    elif parser == PDFParser.pdf_miner:
        loader = PDFMinerLoader(str(paper))
    elif parser == PDFParser.py_mu_pdf:
        loader = PyMuPDFLoader(str(paper))
    elif parser == PDFParser.py_pdf:
        loader = PyPDFLoader(str(paper))
    elif parser == PDFParser.pdfplumber:
        loader = PDFPlumberLoader(str(paper))
    elif parser == PDFParser.unstructured:
        loader = init_unstructured_loader(open(str(paper), "rb"), include_page_breaks, mode, pdf_infer_table_structure,
                                          strategy)
    else:
        loader = PDFPlumberLoader(str(paper))
    where = paper.parent if folder is None else folder / paper.parent.name if recreate_parent else folder
    where.mkdir(parents=True, exist_ok=True)
    docs: list[Document] = loader.load()
    upd_where: Path = (where / paper.stem) if subfolder else where
    upd_where.mkdir(exist_ok=True)
    if upd_where.exists() and subfolder and do_not_reparse and files(upd_where).len() > 0:
        logger.info(f"avoiding re-parsing, providing result {upd_where}")
        return [upd_where]
    if len(docs) == 1:
        return [upd_where] if subfolder else [write_parsed(docs[0], paper, upd_where, cleaning)]
    else:
        acc = []
        for i, doc in enumerate(docs):
            f = write_parsed(doc, paper, upd_where, cleaning, i)
            acc.append(f)
        return [upd_where] if subfolder else acc


def write_parsed(doc: Document, paper: Path, where: Path, cleaning: bool = False, i: int = -1,
                 logger: Optional["loguru.Logger"] = None):
    if logger is None:
        logger = loguru.logger
    name = f"{paper.stem}_{i}.txt" if i >= 0 else f"{paper.stem}.txt"
    f = (where / name)
    logger.trace(f"writing {f}")
    text = clean_text(doc.page_content) if cleaning else doc.page_content
    f.write_text(text)
    return f


def init_unstructured_loader(bin_file, include_page_breaks, mode, pdf_infer_table_structure, strategy: str):
    return UnstructuredPDFLoader(file_path=None, file=bin_file, mode=mode,
                                 pdf_infer_table_structure=pdf_infer_table_structure,
                                 strategy=strategy, include_page_breaks=include_page_breaks)


def try_parse_paper(paper: Path, folder: Optional[Path] = None,
                    parser: PDFParser = PDFParser.unstructured,
                    recreate_parent: bool = False,
                    subfolder: bool = True,
                    do_not_reparse: bool = True,
                    cleaning: bool = False,
                    mode: str = "single", strategy: str = "auto",
                    pdf_infer_table_structure: bool = True,
                    include_page_breaks: bool = False,
                    logger: Optional["loguru.Logger"] = None
                    ) -> Try[List[Path]]:
    return Try.of(
        lambda: parse_paper(paper, folder, parser, recreate_parent, cleaning, subfolder, do_not_reparse, mode, strategy,
                            pdf_infer_table_structure, include_page_breaks, logger=logger))


def parse_papers(parse_folder: Path, destination: Optional[Path] = None,
                 parser: PDFParser = PDFParser.py_mu_pdf,
                 recreate_parent: bool = False, cores: Optional[int] = None,
                 subfolder: bool = True,
                 do_not_reparse: bool = True,
                 cleaning: bool = False,
                 mode: str = "single", strategy: str = "auto",
                 pdf_infer_table_structure: bool = True,
                 include_page_breaks: bool = False,
                 logger: Optional["loguru.Logger"] = None
                 ):
    """
    Function to parse multiple papers using multiple cores.
    The function employs multiprocessing to speed up the process.

    Args:
        parse_folder (Path): The folder where the papers (PDF files) are located.
        destination (Optional[Path]): The destination folder where parsed papers will be saved.
                                      If not provided, the parsed papers are saved in their original folder.
        parser (PDFParser): which pdf parsing library to choose
        mode (str): The mode to parse the papers. Default is "single".
        strategy (str): The strategy to parse the papers. Default is "auto".
        pdf_infer_table_structure (bool): If True, attempts to infer table structure in PDFs. Default is True.
        include_page_breaks (bool): If True, includes page breaks in the parsed output. Default is False.
        recreate_parent (bool): If True, recreates the parent directory structure in the destination folder. Default is False.
        cores (Optional[int]): The number of cores to use. If not provided, uses all available cores.

    Returns:
        list[Path], list[Failure]: A list of paths to the parsed papers and a list of Failtures
    """
    if logger is None:
        logger = loguru.logger
    papers: list[Path] = traverse(parse_folder, lambda p: "pdf" in p.suffix)
    logger.info(f"indexing {len(papers)} papers")
    parsed = []
    errors = []

    cores = cpu_count() if cores is None else min(cpu_count(), cores)

    with Pool(cores) as p:
        parse_func = partial(try_parse_paper, folder=destination, parser=parser,
                             recreate_parent=recreate_parent, subfolder=subfolder, do_not_reparse=do_not_reparse,
                             cleaning=cleaning,
                             mode=mode, strategy=strategy,
                             pdf_infer_table_structure=pdf_infer_table_structure,
                             include_page_breaks=include_page_breaks, logger=logger
                             )
        results = p.map(parse_func, papers)
        for result in results:
            if isinstance(result, pynction.monads.try_monad.Success):
                parsed = parsed + result._value
            elif isinstance(result, pynction.monads.try_monad.Failure):
                errors.append(result._e)
            else:
                logger.warning(f"unpredicted type of the {result}")

    logger.info("papers parsing finished!")
    if len(errors) > 0:
        logger.warning(f"errors discovered: {errors}")
    return results, errors


@click.group(invoke_without_command=False)
@click.pass_context
def app(ctx: Context):
    # if ctx.invoked_subcommand is None:
    #    click.echo('Running the default command...')
    #    test_index()
    pass


@app.command("count_tokens")
@click.option('--path', type=click.Path(exists=True), help="folder to parse papers in")
@click.option('--model', default='gpt-3.5-turbo-16k', help='model to use, gpt-3.5-turbo-16k by default')
@click.option("--suffix", default=".txt", help="suffix in the files to evaluate, .txt by default")
@click.option("--price", type=click.FLOAT, default=0.0001, help="price for 1K tokens")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value,
              help="logging level")
def count_tokens_command(path: Path, model: str, suffix: str, price: float, log_level: str):
    from loguru import logger
    if log_level.upper() != LogLevel.NONE.value:
        logger.add(sys.stdout, level=log_level.upper())
    where = Path(path)
    if where.is_dir():
        papers: list[Path] = traverse(where, lambda p: suffix in p.name)
        tokens_price = seq(papers).map(lambda p: num_tokens_openai(p.read_text(encoding="utf-8"), model, price))
        num = tokens_price.map(lambda r: r[0])
        money = tokens_price.map(lambda r: r[1])
        num_sum = num.sum()
        money_sum = money.sum()
        num_avg = num.average()
        money_avg = money.average()
        num_max = num.max()
        money_max = money.max()
        logger.info(f"Checked {len(papers)} papers. TOTAL TOKENS = {num_sum} , COST = {money_sum}")
        logger.info(
            f"PER PAPER: \n average tokens {num_avg} , cost {money_avg}\n max tokens = {num_max} , max cost = {money_max}")
        return num, money
    else:
        content = path.read_text(encoding="utf-8")
        return num_tokens_openai(content, model)


def try_download_and_parse(doi: str,
                           destination: Path,
                           selenium_on_fail: bool = False,
                           scihub_on_fail: bool = False,
                           parser: PDFParser = PDFParser.py_mu_pdf.value,
                           subfolder: bool = True,
                           do_not_reparse: bool = True,
                           cleaning: bool = False,
                           mode: str = "single",
                           strategy: str = "fast",
                           infer_tables: bool = True,
                           include_page_breaks: bool = False,
                           recreate_parent: bool = True,
                           selenium_headless: bool=True, selenium_min_wait: int=15, selenium_max_wait: int=60,
                           unpaywall_email: Optional[str] = None,
                           logger: Optional["loguru.Logger"] = None) -> Try[PaperDownload]:
    """
    :param doi: paper's doi
    :param destination:
    :param selenium_on_fail: if use selenium on download failure
    :param scihub_on_fail:
    :param parser:
    :param subfolder:
    :param do_not_reparse:
    :param cleaning:
    :param mode:
    :param strategy:
    :param infer_tables:
    :param include_page_breaks:
    :param recreate_parent:
    :param selenium_headless:
    :param selenium_min_wait:
    :param selenium_max_wait:
    :param logger:
    :return:
    """
    if logger is None:
        logger = loguru.logger
    result: Try[PaperDownload] = try_download(doi, destination, skip_if_exist=True,
                                              scihub_on_fail=scihub_on_fail,
                                              selenium_on_fail=selenium_on_fail, selenium_headless=selenium_headless,
                                              selenium_min_wait=selenium_min_wait, selenium_max_wait=selenium_max_wait,
                                              unpaywall_email=unpaywall_email,
                                              logger=logger)
    result.on_failure(lambda ex: logger.error(f"Could not resolve the paper {doi} with the following error {str(ex)}"))
    return result.map(lambda r: r.with_parsed(parse_paper(r.pdf, None, parser, recreate_parent,
                                                          subfolder, do_not_reparse, cleaning,
                                                          mode, strategy, infer_tables, include_page_breaks)))


@app.command("download_and_parse")
@click.option('--doi', type=click.STRING, required=True, help="doi of the paper to parse")
@click.option('--folder', type=click.STRING, default=".", help="destination folder")
@click.option('--selenium_on_fail', type=click.BOOL, default=False, help="use selenium for cases when it fails")
@click.option('--scihub_on_fail', type=click.BOOL, default=False,
              help="if schihub should be used as backup resolver. Use it at your own risk and responsibility (false by default)")
@click.option('--parser', type=click.Choice([loader.value for loader in PDFParser]), default=PDFParser.py_mu_pdf.value,
              help="pdf parser to choose from, unstructured by default")
@click.option('--subfolder', type=click.BOOL, default=True, help="if it should create a folder per paper")
@click.option('--do_not_reparse', type=click.BOOL, default=True, help="if we should avoid reparsing")
@click.option('--cleaning', type=click.BOOL, default=False, help="if we should use basic cleaning for the text")
@click.option('--mode', type=click.Choice(["single", "elements", "paged"]), default="single",
              help="paper mode to be used")
@click.option('--strategy', type=click.Choice(["auto", "hi_res", "fast"]), default="fast",
              help="parsing strategy to be used, auto by default, unstructured parser specific")
@click.option('--infer_tables', type=click.BOOL, default=True,
              help="if the table structure should be inferred, unstructured parser specific")
@click.option('--include_page_breaks', type=click.BOOL, default=False,
              help="if page breaks should be included, unstructured parser specific")
@click.option('--recreate_parent', type=click.BOOL, default=False,
              help="if parent folder should be recreated in the new destination")
@click.option('--selenium_headless', type=click.BOOL, default=True, help="run selenium in the headless mode")
@click.option('--selenium_min_wait', type=click.INT, default=15, help="Min selenium timeout")
@click.option('--selenium_max_wait', type=click.INT, default=60, help="Max selenium timeout")
@click.option('--unpaywall_email', type=click.STRING, default="antonkulaga@gmail.com", help="Unpaywall email to use for pdf resolution")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value,
              help="logging level")
def download_and_parse_command(doi: str, folder: str, selenium_on_fail: bool, scihub_on_fail: bool, parser: str,
                               subfolder: bool, do_not_reparse: bool,
                               cleaning: bool, mode: str, strategy: str, infer_tables: bool,
                               include_page_breaks: bool, recreate_parent: bool,
                               selenium_headless: bool, selenium_min_wait: int, selenium_max_wait: int,
                               unpaywall_email: Optional[str],
                               log_level: str):
    from loguru import logger
    if log_level.upper() != LogLevel.NONE.value:
        logger.add(sys.stdout, level=log_level.upper())
    results = try_download_and_parse(doi,
                                     Path(folder),
                                     selenium_on_fail,
                                     scihub_on_fail,
                                     PDFParser[parser],
                                     subfolder,
                                     do_not_reparse,
                                     cleaning,
                                     mode,
                                     strategy,
                                     infer_tables,
                                     include_page_breaks,
                                     recreate_parent,
                                     selenium_headless,
                                     selenium_min_wait,
                                     selenium_max_wait,
                                     unpaywall_email,
                                     logger)
    results.on_success(lambda r: logger.info(f"Results of {doi} parsing to {r.parsed}:"))
    results.on_failure(lambda e: logger.error(str(e)))
    return results


@app.command("parse_paper")
@click.option('--paper', type=click.Path(exists=True), help="paper pdf to parse")
@click.option('--destination', type=click.STRING, default=".", help="destination folder")
@click.option('--parser', type=click.Choice([loader.value for loader in PDFParser]),
              default=PDFParser.unstructured.value, help="pdf parser to choose from, unstructured by default")
@click.option('--recreate_parent', type=click.BOOL, default=False,
              help="if parent folder should be recreated in the new destination")
@click.option('--subfolder', type=click.BOOL, default=True, help="if it should create a folder per paper")
@click.option('--do_not_reparse', type=click.BOOL, default=True, help="if we should avoid reparsing")
@click.option('--mode', type=click.Choice(["single", "elements", "paged"]), default="single",
              help="paper mode to be used")
@click.option('--strategy', type=click.Choice(["auto", "hi_res", "fast"]), default="fast",
              help="parsing strategy to be used, auto by default, unstructured parser specific")
@click.option('--infer_tables', type=click.BOOL, default=True,
              help="if the table structure should be inferred, unstructured parser specific")
@click.option('--include_page_breaks', type=click.BOOL, default=False,
              help="if page breaks should be included, unstructured parser specific")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value,
              help="logging level")
def parse_paper_command(paper: str, destination: str, parser: str, recreate_parent: bool, subfolder: bool,
                        do_not_reparse: bool, mode: str, strategy: str, infer_tables: bool, include_page_breaks: bool,
                        log_level: str):
    from loguru import logger
    if log_level.upper() != LogLevel.NONE.value:
        logger.add(sys.stdout, level=log_level.upper())
    paper_file = Path(paper)
    destination_folder = Path(destination)
    logger.info(
        f"parsing paper {paper} with mode={mode} {'' if destination_folder is None else 'destination folder ' + destination}")
    return parse_paper(paper_file, None, PDFParser[parser], recreate_parent, True, subfolder, do_not_reparse, mode,
                       strategy, infer_tables, include_page_breaks)


@app.command("parse_folder")
@click.option('--folder', type=click.Path(exists=True), help="folder to parse papers in")
@click.option('--destination', type=click.STRING, default=None, help="destination folder")
@click.option('--parser', type=click.Choice([loader.value for loader in PDFParser]),
              default=PDFParser.unstructured.value, help="pdf parser to choose from, unstructured by default")
@click.option('--mode', type=click.Choice(["single", "elements", "paged"]), default="single",
              help="paper mode to be used")
@click.option('--strategy', type=click.Choice(["auto", "hi_res", "fast"]), default="fast",
              help="parsing strategy to be used, auto by default")
@click.option('--infer_tables', type=click.BOOL, default=True, help="if the table structure should be inferred")
@click.option('--include_page_breaks', type=click.BOOL, default=False, help="if page breaks should be included")
@click.option('--cores', '-t', type=int, default=None, help='Number of cores to use')
@click.option('--recreate_parent', type=click.BOOL, default=False,
              help="if parent folder should be recreated in the new destination")
@click.option('--cleaning', type=click.BOOL, default=True, help="if we should use basic cleaning for the text")
@click.option('--subfolder', type=click.BOOL, default=True, help="if it should create a folder per paper")
@click.option('--do_not_reparse', type=click.BOOL, default=True, help="if we should avoid reparsing")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value,
              help="logging level")
def parse_folder_command(folder: str, destination: str, parser: str, mode: str, strategy: str, infer_tables: bool,
                         include_page_breaks: bool, cores: Optional[int],
                         recreate_parent: bool, cleaning: bool,
                         subfolder: bool, do_not_reparse: bool, log_level: str):
    from loguru import logger
    if log_level.upper() != LogLevel.NONE.value:
        logger.add(sys.stdout, level=log_level.upper())
    parse_folder = Path(folder)
    destination_folder = Path(destination) if destination is not None else None
    logger.info(
        f"parsing paper {folder} with mode={mode} {'' if destination_folder is None else 'destination folder ' + destination}")
    return parse_papers(parse_folder, destination_folder, PDFParser[parser], recreate_parent, cores, subfolder,
                        do_not_reparse, cleaning, mode, strategy, infer_tables, include_page_breaks, logger)


if __name__ == '__main__':
    app()



#!/usr/bin/env python3

from __future__ import annotations
import asyncio
import json
import os
import xml.etree.ElementTree as ET
from collections import OrderedDict
from concurrent.futures import Executor
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Optional, List
import click
import loguru
import requests
from click import Context
from functional import seq
from pynction import Try
from scidownl import scihub_download
from semanticscholar import SemanticScholar
from semanticscholar.Paper import Paper
from dataclasses import dataclass
import sys

from pycomfort.config import LOG_LEVELS, configure_logger, LogLevel
from unpywall import Unpywall
from unpywall.utils import UnpywallCredentials

#DownloadedPaper = (str, Optional[Path], Optional[Path]) #type synonim for doi, Path, Path of the downloaded paper
@dataclass
class PaperDownload:
    id: str
    pdf: Optional[Path]
    metadata: Optional[Path]
    parsed: Optional[list[Path]] = None
    url: Optional[str] = None

    def get_paper_folder(self) -> Optional[Path]:
        if self.pdf is None:
            return None
        else:
            return self.pdf.parent / self.pdf.stem

    def with_pdf(self, pdf: Path):
        self.pdf = pdf
        return self

    def with_url(self, url: str):
        self.url = url
        return self

    def with_parsed(self, parsed: Optional[list[Path]]):
        self.parsed = parsed
        return self

def _pdf_path_for_doi(doi: str, folder: Path, name: Optional[str] = None, create_parent: bool = True) -> Path:
    result = (folder / f"{doi}.pdf").absolute().resolve() if name is None else (folder / f"{name.replace('.pdf', '')}.pdf").abewsw1wswqssolute().resolve()
    if create_parent:
        if not result.parent.exists():
            result.parent.mkdir(exist_ok=True, parents=True)
    return result

def schihub_doi(doi: str, paper: Path, meta: Optional[Path] = None, logger: Optional["loguru.Logger"] = None) -> PaperDownload:
    """
    If you use scihub you should know that it can resolve both openaccess and paywalled articles.
    If you download paywalled articles it can be illegal in some of the countries.
    We are not responsible for how you are using the software, so do it at your own risk.
    :param doi: papers doi
    :param paper:
    :param meta:
    :return:
    """
    if logger is None:
        logger = loguru.logger
    doi_url = f"https://doi.org/{doi}"
    scihub_download(doi_url, paper_type="doi", out=str(paper))
    if paper.exists():
        logger.info(f"downloaded {doi_url} to {paper}")
    return PaperDownload(doi, paper, meta)


#@logger.catch(reraise=False)
def doi_from_pubmed(pubmed_id: str):
    """
    resolves doi by pubmed id
    :param pubmed_id:
    :return:
    """
    base_url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"
    params = {
        "db": "pubmed",
        "id": pubmed_id,
        "retmode": "xml"
    }
    response = requests.get(base_url, params=params)
    #response.raise_for_status()
    root = ET.fromstring(response.content)
    article = root.find(".//PubmedArticle")
    if article is None:
        raise ValueError("PubMed ID not found")
    doi_element = article.find(".//ArticleId[@IdType='doi']")
    if doi_element is None:
        raise ValueError("DOI not found for this PubMed ID")
    return doi_element.text


def try_doi_from_pubmed(pubmed: str) -> Try[str]:
    """
    doi_from_pubmed wrapped in try
    :param pubmed:
    :return:
    """
    return Try.of(lambda: doi_from_pubmed(pubmed))


def resolve_semantic_scholar(paper_id: str,
                              metadata: Optional[Path] = None,
                              logger: Optional["loguru.Logger"] = None) -> PaperDownload:
    if logger is None:
        logger = loguru.logger
    logger.info(f"RESOLVING PAPER {paper_id} with meta and {metadata}")
    sch = SemanticScholar()
    paper: Paper = sch.get_paper(paper_id)
    if metadata is not None:
        json_data = json.dumps(paper.raw_data)
        metadata.touch(exist_ok=True)
        metadata.write_text(json_data)
        logger.info(f"metadata for {paper_id} successfully written to {metadata}")
    if paper.openAccessPdf is not None and "url" in paper.openAccessPdf:
        url = paper.openAccessPdf["url"]
    else:
        url = None
    return PaperDownload(paper_id, None, metadata, None, url)

#@logger.catch(reraise=False)
def download_semantic_scholar(paper_id: str,
                              download: Path,
                              metadata: Optional[Path] = None,
                              raise_on_no_pdf: bool = False,
                              headers: Optional[dict] = None,
                              logger: Optional["loguru.Logger"] = None) -> PaperDownload:
    if logger is None:
        logger = loguru.logger
    sch = SemanticScholar()
    paper: Paper = sch.get_paper(paper_id)
    if metadata is not None:
        json_data = json.dumps(paper.raw_data)
        metadata.touch(exist_ok=True)
        metadata.write_text(json_data)
        logger.info(f"metadata for {paper_id} successfully written to {metadata}")
    if paper.openAccessPdf is not None and "url" in paper.openAccessPdf:
        url = paper.openAccessPdf["url"]
        if download is not None:
            simple_download(url, download, headers, logger)
        else:
            message = f"could not find open-access pdf for {paper_id}"
            if raise_on_no_pdf:
                raise Exception(message)
            else:
                logger.info(message)
                return PaperDownload(paper_id, download, metadata, url = url)
    return PaperDownload(paper_id, download, metadata)


def simple_download(url: str, download: Path, headers: Optional[dict] = None, logger: Optional["loguru.Logger"] = None):
    if headers is None:
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3'
        }
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    download.touch(exist_ok=True)
    download.write_bytes(response.content)
    logger.info(f"downloading open-access pdf for {url} successfully downloaded to {download}")
    return download


def try_download(doi: str,
                 destination: Path,
                 skip_if_exist: bool = True,
                 name: Optional[str] = None,
                 scihub_on_fail: bool = False,
                 unpaywall_email: Optional[str] = None,
                 selenium_on_fail: bool = False,
                 selenium_headless: bool=True, selenium_min_wait: int=12, selenium_max_wait: int=60,
                 logger: Optional["loguru.Logger"] = None
                 ) -> Try[PaperDownload]:
    """
    :param doi:
    :param destination:
    :param skip_if_exist:
    :param name:
    :param scihub_on_fail: use sci-hub on fail. Note: use it at your own risk. If the paper you download is not open-access in some countries and cases sci-hub use will not be legal
    :param unpaywall_email:
    :param selenium_on_fail:
    :param selenium_headless:
    :param selenium_min_wait:
    :param selenium_max_wait:
    :param logger:
    :return:
    """
    if logger is None:
        logger = loguru.logger
    doi_prefix = "https://doi.org/"
    # for example from https://doi.org/10.1016/j.stemcr.2022.09.009
    doi = doi.replace(doi_prefix, "")
    paper = _pdf_path_for_doi(doi, destination, name, True)
    meta = paper.parent / paper.name.replace(".pdf", "_meta.json")
    if skip_if_exist and paper.exists():
        p = PaperDownload(doi, paper, meta)
        if p.parsed is None and p.get_paper_folder() is not None:
            p.parsed = [p.get_paper_folder()]
        logger.info(f"paper {paper} already exists, skipping!")
        return Try.of(lambda: p)

    p: PaperDownload = resolve_semantic_scholar(doi, meta, logger)
    if unpaywall_email is not None:
        unpaywall_email = os.getenv("UNPAYWALL_EMAIL")
    if unpaywall_email is not None and p.url is None:
        UnpywallCredentials(unpaywall_email)
        p.url = Unpywall.get_pdf_link(doi)
    try_simple = Try.of(lambda: p.with_pdf(simple_download(p.url, paper, logger = logger)))
    if selenium_on_fail:
        from getpaper.selenium_download import download_pdf_selenium
        logger.info(f"paper to deal with is {p}")
        before_last = try_simple.catch(lambda ex: p.with_pdf(download_pdf_selenium(p.url, destination, selenium_headless, selenium_min_wait, selenium_max_wait, final_path=paper, logger=logger)))
    else:
        before_last = try_simple
    before_last.on_failure(lambda e: logger.error(e))
    return before_last.catch(lambda _: schihub_doi(doi, paper, meta, logger)) if scihub_on_fail else before_last


def download_pubmed(pubmed: str, destination: Path, skip_if_exist: bool = True, name: Optional[str] = None, scihub_on_fail: bool = False):
    """
    downloads paper by its pubmed id
    :param pubmed: pubmed id
    :param destination: where to store the result
    :param skip_if_exist:
    :param name:
    :param scihub_on_fail: if SciHub should be used as back up resolver. False by default. For paywalled articles it can be illegal in some of the countries, so use it at your own risk.
    :return:
    """
    try_resolve = try_doi_from_pubmed(pubmed)
    return try_resolve.flat_map(lambda doi: try_download(doi, destination, skip_if_exist, name, scihub_on_fail))


async def download_async(executor: Executor,
                             doi: str, destination: Path,
                             skip_if_exist: bool = True,
                             name: Optional[str] = None,
                             scihub_on_fail: bool = False,
                             logger: Optional["loguru.Logger"] = None
                         ) -> PaperDownload:
    """
    Asynchronously download a paper using its DOI.

    Args:
        executor (Executor): The ThreadPoolExecutor to run blocking IO in.
        doi (str): The DOI of the paper to download.
        destination (Path): The directory where the downloaded paper should be stored.
        skip_if_exist (bool): If True, skip the download if the paper already exists in the destination. Default is True.
        name (Optional[str]): The name of the file to save the paper as. If not provided, use the DOI. Default is None.
        param scihub_on_fail: if SciHub should be used as back up resolver. False by default. For paywalled articles it can be illegal in some of the countries, so use it at your own risk.


    Returns:
        (str, Path): A tuple containing the DOI of the paper and the path to the downloaded paper.
    """
    if logger is None:
        logger = loguru.logger

    paper = _pdf_path_for_doi(doi, destination, name, True)
    meta = paper.parent / paper.name.replace(".pdf", "_meta.json")

    # Get a reference to the current event loop
    loop = asyncio.get_event_loop()

    _ = await loop.run_in_executor(executor, lambda: try_download(doi, destination, skip_if_exist, name, scihub_on_fail, logger))

    # Return the DOI and path to the downloaded paper
    return PaperDownload(doi, paper, meta)


def download_papers(dois: List[str], destination: Path, threads: int, scihub_on_fail: bool = False) -> (OrderedDict[str, PaperDownload], List[str]):
    """
    :param dois: List of DOIs of the papers to download
    :param destination: Directory where to put the downloaded papers
    :param threads: Maximum number of concurrent downloads
    :param scihub_on_fail: if SciHub should be used as back up resolver. False by default. For paywalled articles it can be illegal in some of the countries, so use it at your own risk.
    :return: tuple with OrderedDict of succeeded results and list of failed dois)
    """
    # Create a ThreadPoolExecutor with desired number of threads
    with ThreadPoolExecutor(max_workers=threads) as executor:
        # Create a coroutine for each download
        coroutines = [download_async(executor, doi, destination, scihub_on_fail=scihub_on_fail) for doi in dois]
        #TODO: can be problematic
        # Get the current event loop, run the downloads, and wait for all of them to finish
        loop = asyncio.get_event_loop()
        downloaded: List[PaperDownload] = loop.run_until_complete(asyncio.gather(*coroutines))
    partitions: List[List[PaperDownload]] = seq(downloaded).partition(lambda kv: isinstance(kv.pdf, Path)).to_list()
    return OrderedDict([(d.id, d) for d in partitions[0]]), [kv.pdf for kv in partitions[1]]


@click.group(invoke_without_command=False)
@click.pass_context
def app(ctx: Context):
    #if ctx.invoked_subcommand is None:
    #    click.echo('Running the default command...')
    #    test_index()
    pass


@app.command("download_semantic_scholar")
@click.option('--doi', type=click.STRING, help="download doi or other paper id")
@click.option('--folder', type=click.Path(), default=".", help="where to download the paper")
@click.option('--skip_existing', type=click.BOOL, default=True, help="if it should skip downloading if the paper exists")
@click.option('--name', type=click.STRING, default=None, help="custom name, used doi of none")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value, help="logging level")
def download_semantic_scholar_command(doi: str, folder: str, skip_existing: bool = True, name: Optional[str] = None, log_level: str = LogLevel.DEBUG.value):
    from loguru import logger
    if log_level.upper() != LogLevel.NONE.value:
        logger.add(sys.stdout, level=log_level.upper())
    logger.info(f"downloading {doi} to {folder}")
    where = Path(folder)
    where.mkdir(exist_ok=True, parents=True)
    paper = _pdf_path_for_doi(doi, where, name, True)
    meta = paper.parent / paper.name.replace(".pdf", "_meta.json")
    if skip_existing and paper.exists():
        if not meta.exists():
            logger.info(f"paper {paper} pdf already exists, however metadata {meta} does not, trying to download only metadata!")
            return download_semantic_scholar(doi, paper, meta)
        else:
            logger.info(f"paper {paper} already exists, skipping!")
            return doi, paper, meta
    else:
        return download_semantic_scholar(doi, paper, meta)


@app.command("download_doi")
@click.option('--doi', required=True, type=click.STRING, help="download doi")
@click.option('--folder', type=click.Path(), default=".", help="where to download the paper")
@click.option('--skip_existing', type=click.BOOL, default=True, help="if it should skip downloading if the paper exists")
@click.option('--name', type=click.STRING, default=None, help="custom name, used doi of none")
@click.option('--selenium_on_fail', type=click.BOOL, default=False, help="use selenium for cases when it fails")
@click.option('--scihub_on_fail', type=click.BOOL, default=False, help="if schihub should be used as backup resolver. Use it at your own risk and responsibility (false by default)")
@click.option('--unpaywall_email', type=click.STRING, default="antonkulaga@gmail.com", help="Unpaywall email to use for pdf resolution")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value, help="logging level")
def download_doi_command(doi: str, folder: str, skip_existing: bool = True, name: Optional[str] = None, selenium_on_fail: bool = False, scihub_on_fail: bool = False, log_level: str = "NONE") -> Try:
    from loguru import logger
    if log_level.upper() != LogLevel.NONE.value:
        logger.add(sys.stdout, level=log_level.upper())
    logger.debug(f"downloading {doi} to {folder}")
    where = Path(folder)
    where.mkdir(exist_ok=True, parents=True)
    result = try_download(doi, where, skip_existing, name, selenium_on_fail=selenium_on_fail, scihub_on_fail=scihub_on_fail)
    result.on_success(lambda p: logger.info(f"successfully downloaded {p.id} from {p.url} to {p.pdf}"))
    return result


@app.command("download_pubmed")
@click.option('--pubmed', type=click.STRING, help="download doi")
@click.option('--folder', type=click.Path(), default=".", help="where to download the paper")
@click.option('--skip_existing', type=click.BOOL, default=True, help="if it should skip downloading if the paper exists")
@click.option('--scihub_on_fail', type=click.BOOL, default=False, help="if schihub should be used as backup resolver. Use it at your own risk and responsibility (false by default)")
@click.option('--name', type=click.STRING, default=None, help="custom name, uses doi if none and pubmed id if pmid")
def download_pubmed_command(pubmed: str, folder: str, skip_existing: bool, name: Optional[str], scihub_on_fail: bool = False):
    where = Path(folder)
    where.mkdir(exist_ok=True, parents=True)
    custom_name = pubmed if name == "pmid" or name == "PMID" else name
    return download_pubmed(pubmed, where, skip_existing, custom_name, scihub_on_fail=scihub_on_fail)


def get_access(
        sch: SemanticScholar,
        paper_ids: List[str],
        fields: Optional[List[str]] = None
) -> List[dict]:
    if fields is None:
        fields = ["externalIds", "isOpenAccess", "openAccessPdf", "title", "year"]
    url = f'{sch.api_url}/paper/batch'
    fields = ','.join(fields)
    parameters = f'&fields={fields}'
    payload = { "ids": paper_ids }
    data = sch._requester.get_data(url, parameters, sch.auth_header, payload)
    return [item for item in data if item is not None]


def check_access(dois: List[str], logger: Optional["loguru.Logger"] = None) ->(List[Paper], List[Paper]):
    if logger is None:
        logger = loguru.logger
    sch = SemanticScholar()
    papers_list = get_access(sch, dois)
    papers = seq(papers_list)
    result = papers.partition(lambda p: "openAccessPdf" in p and p["openAccessPdf"] is not None and "url" in p["openAccessPdf"])
    opened = result[0].to_list()
    closed = result[1].to_list()
    failed = len(dois) - len(papers_list)
    logger.info(f"{len(result[0].to_list())} papers out of {len(dois)}{' out of which ' + str(failed) + ' were not found' if failed >0 else ''}")
    return opened, closed


@app.command("access")
@click.option('--dois', multiple=True)
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default="debug", help="logging level")
def check_access_command(dois: List[str],  log_level: str):
    configure_logger(log_level)
    return check_access(dois)


@app.command("download_papers")
@click.option('--dois', multiple=True)
@click.option('--folder', type=click.Path(), default=".", help="where to download the paper")
@click.option('--threads', '-t', type=int, default=5, help='Number of threads (default: 5)')
@click.option('--scihub_on_fail', type=click.BOOL, default=False, help="if schihub should be used as backup resolver. Use it at your own risk and responsibility (false by default)")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default="debug", help="logging level")
def download_papers_command(dois: List[str], folder: str, threads: int, scihub_on_fail: bool, log_level: str):
    """Downloads papers with the given DOIs to the specified destination."""
    from loguru import logger
    if log_level.upper() != LogLevel.NONE.value:
        logger.add(sys.stdout, level=log_level.upper())
    if not dois:
        dois = []
    # Call the actual function with the provided arguments
    where = Path(folder)
    where.mkdir(exist_ok=True, parents=True)
    return download_papers(dois, where, threads, scihub_on_fail)


if __name__ == '__main__':
    app()


#!/usr/bin/env python3

from __future__ import annotations

import sys

import loguru
from click import Context

try:
    from selenium import webdriver
    from selenium.webdriver.firefox.options import Options as FirefoxOptions
    from selenium.common.exceptions import TimeoutException
    # Use optional_dependency in your code
except ImportError:
    # Handle the case where the optional_dependency is not installed
    print("Optional dependency is not installed. Some features may be unavailable.")

from pathlib import Path
import time
import os
import click
from typing import Optional


def download_pdf_selenium(url: str, download_dir: Path, headless: bool=True, min_wait_time: int=8, max_wait_time: int=60, final_path: Optional[Path] = None, logger: Optional["loguru.Logger"] = None) -> Path:
    if logger is None:
        logger = loguru.logger
    logger.info(f"selenium parameters url {url}, download_dir {download_dir}, headless {headless}, min_wait_time {min_wait_time}, max wait time {max_wait_time}, final path {final_path}")

    download_dir.mkdir(parents=True, exist_ok=True)
    absolute_download_dir = str(download_dir.resolve())

    options = FirefoxOptions()
    if headless:
        options.add_argument("--headless")

    # Set preferences to disable the built-in PDF viewer and auto-download PDFs
    options.set_preference("pdfjs.disabled", True)
    options.set_preference("browser.download.folderList", 2)
    options.set_preference("browser.download.manager.showWhenStarting", False)
    options.set_preference("browser.download.dir", absolute_download_dir)
    options.set_preference("browser.helperApps.neverAsk.saveToDisk", "application/pdf")

    logger.info(f"will use selenium firefox driver to download {url}")
    driver = webdriver.Firefox(options=options)

    try:
        driver.set_page_load_timeout(min_wait_time)
        driver.get(url)
    except TimeoutException:
        logger.info(f"min download time of the {url} timed out but continuing with the selenium download script...")

    try:
        start_time = time.time()
        downloaded_file = None

        while time.time() - start_time < max_wait_time:
            pdf_files = [f for f in download_dir.iterdir() if f.is_file() and f.suffix == '.pdf']

            if pdf_files:
                new_file = max(pdf_files, key=os.path.getctime)
                if downloaded_file != new_file or os.path.getsize(new_file) > 0:
                    downloaded_file = new_file
                    break  # Exit the loop as soon as a new file is found and is not empty

            time.sleep(1)  # Polling interval

        if not downloaded_file:
            raise Exception("Download timed out or no new file detected.")
        return downloaded_file if final_path is None else downloaded_file.rename(final_path)
    except Exception as e:
        raise e
    finally:
        driver.quit()

"""
# Example usage
url = "https://www.cell.com/article/S2213671122004581/pdf"
output_directory = Path("/home/antonkulaga/sources/getpaper/data/output/test/papers")

try:
    downloaded_file_path = download_pdf_selenium(url, output_directory, headless=True)
    print(f"Downloaded file: {downloaded_file_path}")
except Exception as e:
    print(f"An error occurred: {e}")
"""

@click.group(invoke_without_command=False)
@click.pass_context
def app(ctx: Context):
    if ctx.invoked_subcommand is None:
        click.echo('Running the default command...')
        selenium_download_command()
    pass

@app.command("selenium_download")
@click.option('--url', type=click.STRING, required=True, help="url to download")
@click.option('--destination', type=click.Path(exists=True), default="data/output/test/papers", help="folder to parse papers in")
@click.option('--headless', type=click.BOOL, default=True, help="if to run in a headless mode")
@click.option('--min_wait', type=int, default=12, help='Min waiting time')
@click.option('--max_wait', type=int, default=60, help='Max waiting time')
@click.option('--final_path', type=click.Path(), default = None, help="final path")
def selenium_download_command(url: str, destination: str, headless: bool, min_wait: int, max_wait: int, final_path: Optional[Path]):
    logger = loguru.logger
    logger.add(sys.stdout)
    logger.info(f"download {url} to {destination} in headless={headless} mode with min_wait={min_wait} and max_wait={max_wait}")
    result = download_pdf_selenium(url, Path(destination), headless, min_wait, max_wait, final_path=final_path, logger=logger)
    loguru.logger.info(f"downloaded to {result}")
    return result

if __name__ == '__main__':
    app()


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.4.6.1'
DESCRIPTION = 'getpaper - papers download made easy!'
LONG_DESCRIPTION = 'A package with python functions for downloading papers'

# Setting up
setup(
    name="getpaper",
    version=VERSION,
    author="antonkulaga (Anton Kulaga)",
    author_email="<antonkulaga@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        'pycomfort>=0.0.14',
        'click',
        'pynction',
        'langchain>=0.0.338',
        'scidownl>=1.0.2',
        'Deprecated',
        'semanticscholar>=0.5.0',
        'PyMuPDF>=1.23.6',
        "unpywall>=0.2.2"
    ],
    extras_require={
        'selenium': ['selenium', 'webdriver-manager'],
        'unstructured': [
            'pdfminer.six',
            'unstructured>=0.10.25',
            'unstructured-inference>=0.7.9',
            'unstructured[local-inference]',
            'unstructured.PaddleOCR>=2.6.1.3'
            #'git+https://github.com/facebookresearch/detectron2.git'
        ],
        'plumber': ['pdfplumber>=0.10.3']
    },
    keywords=['python', 'utils', 'files', 'papers', 'download'],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    entry_points={
     "console_scripts": [
         "download=getpaper.download:app",
         "parse=getpaper.parse:app"
     ]
    }
)
